unit Evolve1;

(*
    Evolve shows a simple evolutionary model in action.
    Copyright (C) 2007  Robert J. Norton

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, Math, ExtCtrls, XPMan;

type

  { individual living creature }
  TCritter = class(TCollectionItem)
    public
       DNA: string;
       PhenoLabel: TLabel;
       Energy: real;
       Fitness1: real;
       Age: integer;
       Velocity: TPoint;
       function RandSpot(MaxSpot:integer): integer;
       procedure CalcFitness;
       procedure Wander;
       function MutateColor(OldColor: TColor): TColor;
       function MutateDNA(OldDNA: string): string;
       procedure Feed(NewFood: real);
       destructor Destroy; override;
    end;

  { population of all the living critters  }
  { This is the access property for reading each item in the list.
  Notice the use of the default directive.
  This means you can access this property either by its name
     (MyCollection.Items[ThisIndex];)
  or without the name
     (MyCollection[ThisIndex];)}
  TPop = class(TCollection)
  protected
    function GetItem(Index : Integer) : TCritter;
    procedure SetItem(Index : Integer; Value : TCritter);
  public
    AverageFitness1: real;

    function Add(OldCritter: TCritter) : TCritter;
    procedure Seed(NewDNA: string; NewEnergy: real);
    procedure CalcAverageFitness;
    property Items[Index : Integer] : TCritter read GetItem write SetItem; default;
  end;

  { the application class }
  TForm1 = class(TForm)
    Timer1: TTimer;
    GroupBox1: TGroupBox;
    edTarget1: TLabeledEdit;
    AveFitness1: TLabeledEdit;
    XPManifest1: TXPManifest;
    edFood1: TLabeledEdit;
    edPop1: TLabeledEdit;
    edBirths: TLabeledEdit;
    edCycles: TLabeledEdit;
    btPlague: TButton;
    procedure btPlagueClick(Sender: TObject);
    procedure edFood1Change(Sender: TObject);
    procedure edTarget1Change(Sender: TObject);
    procedure Timer1Timer(Sender: TObject);
    procedure FormCreate(Sender: TObject);

  private
    { Private declarations }
    Cycles: longint;
    Births: longint;
    Target1: string;
    TotalFood: integer;                          { total food available to the population }

    procedure DoCycle;

  public
    { Public declarations }
    MyPop: TPop;

  end;

var
  Form1: TForm1;

implementation

uses Evolve2;
const
   COST_OF_LIVING = 1;                           { food units used per cycle to stay alive }
   BIRTH_ENERGY   = 10;                          { gather enough units to make a birth  }
   MUTATION_RATE  = 0.1;                         { 0.1 = 10% will be mutations }
   OLD_AGE        = 30;                          { don't trust any critter over 30 }

{$R *.dfm}

(*
 * Find the Levenshtein distance (LD) between two strings
 * This is a measure of the similarity between two strings,
 * which we will refer to as the source string (s) and the target string (t).
 * The distance is the number of deletions, insertions, or substitutions required to
 * transform s into t.
 *    http://www.merriampark.com/ld.htm
 *)
function LDist(s, t: string): integer;
   function MinOf3(a, b, c: integer): integer;
   begin
      result := Min(Min(a, b), c);
      end;
var
   d: array of array of integer; // matrix
   n: integer;    //length of s
   m: integer;    // length of t
   i: integer;    // iterates through s
   j: integer;    // iterates through t
   s_i: char;     // ith character of s
   t_j: char;     // jth character of t
   cost: integer; // cost
begin
// Step 1
   n := length(s);
   m := length(t);
   if n = 0
   then result := m
   else begin
      if m = 0
      then result := n
      else begin;
         SetLength(d, n + 1, m + 1);
// Step 2
         for i := 0 to n do d[i, 0] := i;
         for j := 0 to m do d[0, j] := j;
// Step 3
         for i := 1 to n do begin
            s_i := s[i];
// Step 4
            for j := 1 to  m do begin
               t_j := t[j];
// Step 5
               if s_i = t_j
               then cost := 0
               else cost := 1;
// Step 6
               d[i,j] := MinOf3(d[i-1, j]+1, d[i, j-1]+1, d[i-1, j-1] + cost);
               end;
            end;
// Step 7
         result := d[n, m];
         end;
      end;
   end;

{ bound an integer to a range }
function Bound(Val, minV, maxV: integer):integer;
begin
   result := max(min(maxV, Val), minV);
   end;

{****************}
{*** TCritter ***}
{****************}

{ reproduce given DNA with possibility of mutation }
function TCritter.MutateDNA(OldDNA: string): string;
var
   idx: integer;

   function RandChar: char;  { random pick on the printable range }
   begin
      result := char(32 + Random(95));           { from ' ' to '~' }
      end;

begin
   if Random < MUTATION_RATE
   then case Random(4) of
      // Add random letter in random location, limit 100
      0: if length(OldDNA) < 100
         then begin
            idx := Random(length(OldDNA) + 1);
            OldDNA := copy(OldDNA, 1, idx) + RandChar + copy(OldDNA, idx + 1, length(OldDNA));
            end;
      // Remove random letter, limit clipping to lenght 1
      1: if length(OldDNA) > 1
         then begin
            idx := 1 + Random(length(OldDNA));
            delete(OldDNA, idx, 1);
            end;
      // Replace random letter into random location
      2: begin
            idx := 1 + Random(length(OldDNA));
            OldDNA[idx] := RandChar;
            end;
      // Change case of letter in random location
      3: begin
            idx := 1 + Random(length(OldDNA));
            if OldDNA[idx] in ['a'..'z', 'A'..'Z']
            then OldDNA[idx] := char(ord(OldDNA[idx]) xor $20);
            end;
      end;
   result := OldDNA;
   end;

{ reproduce color similar to parent's }
function TCritter.MutateColor(OldColor: TColor): TColor;
   function MutateOne(Val: integer): integer;
   begin
      result := Bound(Val + Random(33) - 16, 128, 255);
      end;
var
   Red, Green, Blue: integer;
begin
   Red   :=  OldColor         and $FF;
   Green := (OldColor SHR  8) and $FF;
   Blue  := (OldColor SHR 16) and $FF;
   Red   := MutateOne(  Red);
   Green := MutateOne(Green);
   Blue  := MutateOne( Blue);

   result := (Blue SHL 16) + (Green SHL 8) + Red;
   end;

{ get a random value within a range }
function TCritter.RandSpot(MaxSpot: integer): integer;
begin
   result := 5 + Random(MaxSpot - 10);
   end;

{ find its own fitness level against current target }
procedure TCritter.CalcFitness;
var
   T: string;
begin
   T := Form1.Target1;
   { count 1/2 point for right letter, wrong case. }
   Fitness1 := length(T) -
              ((LDist(T, Self.DNA) + LDist(UpperCase(T), UpperCase(Self.DNA)))/2);
   end;

{ animate labels as though drifting in the tank }
procedure TCritter.Wander;
begin
   Velocity.Y := Bound(Velocity.Y + Random(3) - 1, -5, 5);
   if PhenoLabel.Top < 5
   then Velocity.Y := abs(Velocity.Y)
   else begin
      if PhenoLabel.Top > (Form2.Tank.Height - (PhenoLabel.Height + 5))
      then Velocity.Y := 0 - abs(Velocity.Y)
      end;
   PhenoLabel.Top := PhenoLabel.Top + Velocity.Y;

   Velocity.X := Bound(Velocity.X + Random(3) - 1, -5, 5);
   if PhenoLabel.Left < 5
   then Velocity.X := abs(Velocity.X)
   else begin
      if PhenoLabel.Left > (Form2.Tank.Width - (PhenoLabel.Width + 5))
      then Velocity.X := 0 - abs(Velocity.X);
      end;
   PhenoLabel.Left := PhenoLabel.Left + Velocity.X;
   end;

{ feeding per cycle, includes cost of living }
procedure TCritter.Feed(NewFood: real);
begin
   Energy := Energy + NewFood;
   Age := Age + 1;
   end;

{ be sure to free critter's label on destruction }
destructor TCritter.Destroy;
begin
   Phenolabel.Parent := nil;                     { pull it from the tank }
   Phenolabel.Free;
   inherited;
   end;

{************}
{*** TPop ***}
{************}

{ Create a critter manually }
procedure TPop.Seed(NewDNA: string; NewEnergy: real);
var
   Temp: TCritter;
begin
   Temp := TCritter(inherited Add);                 { adding to collection creates a critter }
   with Temp do begin
      DNA := NewDNA;
      PhenoLabel := TLabel.Create(nil);               { remember to explicitly free }
      PhenoLabel.Caption := NewDNA;
      PhenoLabel.Parent := Form2.Tank;
      PhenoLabel.Top := RandSpot(Form2.Tank.Height - PhenoLabel.Height);
      PhenoLabel.Left := RandSpot(Form2.Tank.Width - PhenoLabel.Width);
      PhenoLabel.Font.Color := clSkyBlue;
      PhenoLabel.Font.Size := 10;
      PhenoLabel.Font.Style := [fsBold];
      CalcFitness;
      Age := 0;
      Energy := NewEnergy;
      Velocity.X := 0;
      Velocity.Y := 0;
      end;
   end;

{ This is the method that adds a child critter to the population }
function TPop.Add(OldCritter: TCritter): TCritter;
begin
   Result := TCritter(inherited Add);                 { adding to collection creates a critter }
   with Result do begin                               { pattern after parrten }
      DNA := MutateDNA(OldCritter.DNA);
      PhenoLabel := TLabel.Create(nil);               { remember to explicitly free }
      PhenoLabel.Caption := Result.DNA;
      PhenoLabel.Parent := Form2.Tank;
      if OldCritter.PhenoLabel.Top < (Form2.Tank.Height - (5 + (2 * OldCritter.PhenoLabel.Height)))
      then PhenoLabel.Top := OldCritter.PhenoLabel.Top + OldCritter.PhenoLabel.Height
      else PhenoLabel.Top := OldCritter.PhenoLabel.Top - OldCritter.PhenoLabel.Height;
      PhenoLabel.Left := OldCritter.PhenoLabel.Left;
      PhenoLabel.Font.Color := MutateColor(OldCritter.PhenoLabel.Font.Color);
      PhenoLabel.Font.Size := OldCritter.PhenoLabel.Font.Size;
      PhenoLabel.Font.Style := OldCritter.PhenoLabel.Font.Style;
      CalcFitness;
      Age := 0;
      Energy := OldCritter.Energy;                    { was cut in half prior }
      Velocity.X := OldCritter.Velocity.X;            { drifting same direction as parent }
      Velocity.Y := OldCritter.Velocity.Y;
      end;
   end;


{ This is the read access method for the Items property. }
function TPop.GetItem(Index: integer): TCritter;
begin
   result := TCritter(inherited GetItem(Index));
   end;

{ find the average fitness in the population }
procedure TPop.CalcAverageFitness;
var
   TotalFitness1: real;
   Eaters1: integer;
   i: integer;
begin
   // find total amount of fitness in the population
   TotalFitness1 := 0;
   Eaters1 := 0;

   for i := 0 to Count - 1 do begin
      TotalFitness1 := TotalFitness1 + Items[i].Fitness1;
      Eaters1 := Eaters1 + 1;
      end;

   if Eaters1 > 0
   then AverageFitness1 := TotalFitness1/Eaters1
   else AverageFitness1 := 0;
   Form1.AveFitness1.Text := FloatToStrF(AverageFitness1, ffFixed, 10, 2);
   if IsIconic(Application.Handle)
   then Application.Title := 'Evolve ' + FloatToStrF(AverageFitness1, ffFixed, 10, 2)
   else Application.Title := 'Evolve';

   end;


{ This is the write method for the Items property}
procedure TPop.SetItem(Index : Integer; Value : TCritter);
begin
   inherited SetItem(Index, Value);
   end;

{**************}
{*** TForm1 ***}
{**************}

{ a year in the life of the critter }
procedure TForm1.DoCycle;
var
   i: integer;
   Food: real;
begin
   MyPop.CalcAverageFitness;

   // check for births
   for i := 0 to MyPop.Count - 1 do begin
      if MyPop[i].Energy >= BIRTH_ENERGY
      then begin
         MyPop[i].Energy := MyPop[i].Energy / 2;           { take energy to make a birth }
         MyPop.Add(MyPop[i]);
         Births := Births + 1;
         MyPop.CalcAverageFitness;
         edBirths.Text := IntToStr(Births);
         end;
      end;

   // feed every critter proportional to its fitness
   for i := 0 to MyPop.Count - 1 do begin
      Food := (TotalFood / MyPop.Count)
              + (MyPop[i].Fitness1 - MyPop.AverageFitness1);
      Food := Min(Food, 2.0);                              { limit food to double rations }
      Food := Food + (Random - 0.5);                       { lucky or not on getting food }
      MyPop[i].Feed(Food - COST_OF_LIVING);
      MyPop[i].Wander;                                     { decorative wiggling }

      edPop1.Text := IntToStr(MyPop.Count);                { results of this feeding }
      end;

   // check for deaths
  for i := MyPop.Count - 1 downto 0 do begin
      if MyPop.Count > 1
      then begin                                           { spare the last critter }
         if MyPop[i].Energy < 0
         then MyPop.Delete(i)                              { starved to death }
         else begin
            if MyPop[i].Age > OLD_AGE
            then begin
               if Random > 0.75
               then MyPop.Delete(i);                       { died of old age }
               end;
            end;
         end;
      end;
   end;

{ user may have changed the goal for the niche }
procedure TForm1.edTarget1Change(Sender: TObject);
var
   i: integer;
begin
   if Target1 <> edTarget1.Text
   then begin
      Target1 := edTarget1.Text;
      if MyPop <> nil
      then for i := 0 to MyPop.Count - 1 do MyPop[i].CalcFitness;
      end;
   end;

{ set up the initial environment }
procedure TForm1.FormCreate(Sender: TObject);
begin
   Randomize;
   Target1 := edTarget1.Text;

   MyPop := TPop.Create(TCritter);

   Cycles := 0;
   Births := 0;
   TotalFood := 80;
   end;

{ create the passage of time }
procedure TForm1.Timer1Timer(Sender: TObject);
begin
   DoCycle;
   Cycles := Cycles + 1;
   edCycles.Text := IntToStr(Cycles);
   end;

{ somebody has changed the food amount }
procedure TForm1.edFood1Change(Sender: TObject);
begin
   TotalFood := StrToIntDef(edFood1.Text, 80);
   end;

{ kill all but 5 critters at tail of list }
procedure TForm1.btPlagueClick(Sender: TObject);
begin
   while MyPop.Count > 5
   do myPop.Delete(0);
   end;

end.
